@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality font engine         */
 /*                                                                          */
-/*  Copyright 2005, 2006 by                                                 */
+/*  Copyright 2005, 2006, 2007 by                                           */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*  ftvalid: Validates layout related tables of OpenType and                */
@@ -38,6 +38,18 @@
 
 
   static char*  execname;
+
+  typedef enum 
+  { 
+    OT_VALIDATE = 0,
+    GX_VALIDATE,
+    CKERN_VALIDATE,
+
+    LAST_VALIDATE,
+
+  } ValidatorType;
+  static ValidatorType  validator;
+
   
   typedef struct  TableSpecRec_
   {
@@ -46,29 +58,9 @@
 
   } TableSpecRec, *TableSpec;
 
-
-  typedef enum 
-  { 
-#define OT_VALIDATOR_SYMBOL "ot" 
-    OT_VALIDATE = 0,
-#define GX_VALIDATOR_SYMBOL "gx" 
-    GX_VALIDATE,
-#define CKERN_VALIDATOR_SYMBOL "ckern" 
-    CKERN_VALIDATE,
-
-  } ValidatorType;
-
-  static const char*  validator_symbols[] = { OT_VALIDATOR_SYMBOL, 
-                                              GX_VALIDATOR_SYMBOL, 
-                                              CKERN_VALIDATOR_SYMBOL,
-                                              NULL };
-
-  static ValidatorType  validator;
-
-  
 #define MAKE_TABLE_SPEC( x ) { TTAG_##x, FT_VALIDATE_##x }
 
-  static const TableSpecRec  ot_table_spec[] = 
+  static TableSpecRec  ot_table_spec[] = 
   {
     MAKE_TABLE_SPEC( BASE ),
     MAKE_TABLE_SPEC( GDEF ),
@@ -78,7 +70,7 @@
   };
 #define N_OT_TABLE_SPEC  ( sizeof ( ot_table_spec ) / sizeof ( TableSpecRec ) )
 
-  static const TableSpecRec  gx_table_spec[] = 
+  static TableSpecRec  gx_table_spec[] = 
   {
     MAKE_TABLE_SPEC( feat ),
     MAKE_TABLE_SPEC( mort ),
@@ -94,6 +86,85 @@
 #define N_GX_TABLE_SPEC  ( sizeof ( gx_table_spec ) / sizeof ( TableSpecRec ) )
 
 
+  typedef struct  ValidatorRec_
+  {
+    ValidatorType  type;
+    const char*    symbol;
+
+    const char*    unimplemented_message;
+    int         (* is_implemented)( FT_Library library );
+
+    FT_Error    (* run)           ( FT_Face     face,
+				    const char* tables,
+				    int         validation_level );
+    int         (* list_tables)   ( FT_Face     face );
+
+    TableSpec       table_spec;   
+    unsigned int    n_table_spec;
+
+  } ValidatorRec, *Validator;
+
+  static int is_ot_validator_implemented    ( FT_Library library );
+  static int is_gx_validator_implemented    ( FT_Library library );
+  static int is_ckern_validator_implemented ( FT_Library library );
+
+  static FT_Error run_ot_validator          ( FT_Face      face, 
+					      const char*  tables, 
+					      int          validation_level );
+  static FT_Error run_gx_validator          ( FT_Face      face, 
+					      const char*  tables, 
+					      int          validation_level );
+  static FT_Error run_ckern_validator       ( FT_Face      face, 
+					      const char*  dialect_request, 
+					      int          validation_level );
+
+  static int list_ot_tables                 ( FT_Face  face );
+  static int list_gx_tables                 ( FT_Face  face );
+  static int list_ckern_tables              ( FT_Face  face );
+
+
+  ValidatorRec validators[] = {
+    {
+      OT_VALIDATE,
+      "ot",
+      ( "FT_OpenType_Validate" 
+	"is disabled. replace FreeType2 with " 
+	"otvalid" 
+	"-enabled version\n" ),
+      is_ot_validator_implemented,
+      run_ot_validator,
+      list_ot_tables,
+      ot_table_spec,
+      N_OT_TABLE_SPEC,
+    },
+    {
+      GX_VALIDATE,
+      "gx",
+      ( "FT_TrueTypeGX_Validate"
+	"is disabled. replace FreeType2 with " 
+	"gxvalid" 
+	"-enabled version\n" ),
+      is_gx_validator_implemented,
+      run_gx_validator,
+      list_gx_tables,
+      gx_table_spec,
+      N_GX_TABLE_SPEC,
+    },
+    {
+      CKERN_VALIDATE,
+      "ckern",
+      ( "FT_ClassicKern_Validate"
+	"is disabled. replace FreeType2 with " 
+	"gxvalid" 		/* NOTE: classic kern validator is in gxvalid. */
+	"-enabled version\n" ),
+      is_ckern_validator_implemented,
+      run_ckern_validator,
+      list_ckern_tables,
+      NULL,
+      0,
+    },
+  };
+
 
   static void
   panic( int          error,
@@ -128,11 +199,22 @@
   }
 
 
+  /* To initialize the internal variable, call this 
+     funtion with FT_Library variable. Then call
+     with NULL. The print messages is printed if
+     call with NULL. */
   static void
-  print_usage( void )
+  print_usage( FT_Library  library_initializer )
   {
-    unsigned int i;
+    unsigned int i, j;
+    Validator v;
+    static FT_Library library;
 
+    if (library_initializer)
+    {
+      library = library_initializer;
+      return ;
+    }
 
     fprintf( stderr, "\n" );
     fprintf( stderr, "ftvalid: layout table validator -- part of the FreeType project\n" );
@@ -141,36 +223,38 @@
     fprintf( stderr, "Usage: %s [options] fontfile\n", execname );
     fprintf( stderr, "\n" );
     fprintf( stderr, "  -t validator              select validator. \n");
-    fprintf( stderr, "                            Currently \"ot\", \"gx\" and \"ckern\" are available.\n" );
+    fprintf( stderr, "                            Available validators: ");
+    for ( i = 0; i < LAST_VALIDATE; i++ )
+    {
+      v = &validators[i];
+      fprintf( stderr, "\"%s\"%s ", v->symbol, v->is_implemented( library )? "": "(NOT IMPLEMENTED)");
+    }
+    fprintf( stderr, "\n");
+
     fprintf( stderr, "\n" );
-    fprintf( stderr, "  -T \"sfnt:tabl:enam:es  \"  [ot or gx] select snft table names to be \n" );
+    fprintf( stderr, "  -T \"sfnt:tabl:enam:es  \"  select snft table names to be \n" );
     fprintf( stderr, "                            validated. `:' is for separating table names.\n" );
     fprintf( stderr, "\n" );
-    fprintf( stderr, "                            Supported tables in ot validator are:\n" );
 
-    fprintf( stderr, "                            " );
-
-    for ( i = 0; i < N_OT_TABLE_SPEC; i++ )
+    for ( i = 0; i < LAST_VALIDATE; i++ )
     {
-      print_tag( stderr, ot_table_spec[i].tag );
-      fprintf( stderr, " " );
+      v = &validators[i];
+
+      if ( v->n_table_spec == 0 )
+	continue;
+
+      fprintf( stderr, "                            Supported tables in %s validator are:\n", v->symbol );
+      fprintf( stderr, "                            " );
+      for ( j = 0; j < v->n_table_spec; j++ )
+      {
+	print_tag( stderr, v->table_spec[j].tag );
+	fprintf( stderr, " " );
+      }
+
+      fprintf( stderr, "\n" );
+      fprintf( stderr, "\n" );      
     }
 
-    fprintf( stderr, "\n" );
-    fprintf( stderr, "\n" );
-    fprintf( stderr, "                            Supported tables in gx validator are:\n" );
-
-    fprintf( stderr, "                            " );
-
-    for ( i = 0; i < N_GX_TABLE_SPEC; i++ )
-    {
-      print_tag( stderr, gx_table_spec[i].tag );
-      fprintf( stderr, " " );
-    }
-
-    fprintf( stderr, "\n" );
-    fprintf( stderr, "\n" );
-
     fprintf( stderr, "  -T \"ms:apple\"             [ckern] select (a) classic kern dialect(s) for \n" );
     fprintf( stderr, "                            validation. `:' is for separating dialect names.\n" );
     fprintf( stderr, "                            If more than one dialects is specified, all\n" );
@@ -248,7 +332,7 @@
     print_tag( stderr, tag );
     fprintf( stderr, "\n" );
 
-    print_usage();
+    print_usage( NULL );
 
     return 0;
   }
@@ -272,7 +356,7 @@
     if (( len % 5 ) != 4 )
     {
       fprintf( stderr, "*** Wrong length of table names\n" );
-      print_usage();
+      print_usage( NULL );
     }
     
     for ( i = 0; i < len; i++ )
@@ -282,7 +366,7 @@
         if ( tables[i] != ':' )
         {
           fprintf( stderr, "*** Wrong table separator: %c\n", tables[i] );
-          print_usage();
+          print_usage( NULL );
         }
         i++;
       }
@@ -372,7 +456,7 @@
                  int                 spec_count )
   {
     printf( "[%s:%s] validation targets: ",
-            execname, validator_symbols[validator] );
+            execname, validators[validator].symbol );
     print_tables( stdout, validation_flags, spec, spec_count );
     printf( "-------------------------------------------------------------------\n" );
   }
@@ -386,28 +470,33 @@
   {
     int  i;
     int  n_passes;
+    int  n_targets;
     
 
-    for ( i = 0, n_passes = 0; i < spec_count; i++ )
+    for ( i = 0, n_passes = 0, n_targets = 0; i < spec_count; i++ )
     {
-      if ( ( spec[i].validation_flag & validation_flags ) &&
-             data[i] != NULL                              )
+      if ( spec[i].validation_flag & validation_flags )
       {
-        printf( "[%s:%s] ", execname, validator_symbols[validator] );
-        print_tag( stdout, spec[i].tag );
-        printf( "...pass\n" );
-        n_passes++;
+	n_targets++;
+	
+	if ( data[i] != NULL )
+	{
+	  printf( "[%s:%s] ", execname, validators[validator].symbol );
+	  print_tag( stdout, spec[i].tag );
+	  printf( "%s", "...pass\n" );
+	  n_passes++;
+	}
       }
     }
 
-    if ( n_passes == 0 )
+    if ( n_passes == 0 && n_targets != 0 )
     {
       printf( "[%s:%s] layout tables are invalid.\n",
-              execname, validator_symbols[validator] );
+              execname, validators[validator].symbol );
       printf( "[%s:%s] set FT2_DEBUG environment variable to \n",
-              execname, validator_symbols[validator] );
+              execname, validators[validator].symbol );
       printf( "[%s:%s] know the validation detail. \n",
-              execname, validator_symbols[validator] );
+              execname, validators[validator].symbol );
     }
   }
 
@@ -416,6 +505,15 @@
    * OpenType related funtions
    */
   static int
+  is_ot_validator_implemented( FT_Library library )
+  {
+    FT_Module mod;
+    
+    mod = FT_Get_Module( library, "otvalid" );
+    return mod? 1: 0;
+  }
+
+  static FT_Error
   run_ot_validator( FT_Face      face,
                     const char*  tables,
                     int          validation_level )
@@ -445,7 +543,7 @@
     for ( i = 0; i < N_OT_TABLE_SPEC; i++ )
       FT_OpenType_Free( face, data[i] );
 
-    return (int)error;
+    return error;
   }
 
 
@@ -466,6 +564,15 @@
    * TrueTypeGX related funtions
    */
   static int
+  is_gx_validator_implemented( FT_Library library )
+  {
+    FT_Module mod;
+    
+    mod = FT_Get_Module( library, "gxvalid" );
+    return mod? 1: 0;
+  }
+
+  static FT_Error
   run_gx_validator( FT_Face      face, 
                     const char*  tables, 
                     int          validation_level )
@@ -495,7 +602,7 @@
     for ( i = 0; i < N_GX_TABLE_SPEC; i++ )
       FT_TrueTypeGX_Free( face, data[i] );
 
-    return (int)error;
+    return error;
   }
 
 
@@ -515,6 +622,16 @@
    * Classic kern related funtions
    */
   static int
+  is_ckern_validator_implemented( FT_Library library )
+  {
+    FT_Module mod;
+
+    mod = FT_Get_Module( library, "gxvalid" );
+    return mod? 1: 0;
+  }
+
+
+  static FT_Error
   run_ckern_validator( FT_Face      face, 
                        const char*  dialect_request, 
                        int          validation_level )
@@ -540,11 +657,11 @@
     else
     {
       fprintf( stderr, "Wrong classic kern dialect: %s\n", dialect_request );
-      print_usage();
+      print_usage( NULL );
     }
 
     printf( "[%s:%s] validation targets: %s...", 
-            execname, validator_symbols[validator], dialect_request );
+            execname, validators[validator].symbol, dialect_request );
 
 
     error = FT_ClassicKern_Validate(
@@ -562,7 +679,7 @@
     
     FT_ClassicKern_Free( face, data );
               
-    return (int)error;
+    return error;
   }
 
   static int
@@ -584,9 +701,12 @@
   main( int     argc, 
         char**  argv )
   {
+    FT_Library  library;
+    FT_Error    error;
+  
     char*  fontfile;
     int    option;
-    int    status;
+
 
     char*  tables;
     int    dump_table_list;
@@ -595,9 +715,17 @@
 #if 0
     int  trace_level;
 #endif  /* 0 */
-    
+
     execname = ft_basename( argv[0] );
 
+    error = FT_Init_FreeType( &library );
+    if ( error )
+      panic ( error, "Could not initialize FreeType library" );
+
+    /* Initialize print_usage internal variable */
+    print_usage( library );
+
+
     /*
      * Parsing options
      */
@@ -621,22 +749,21 @@
       case 't':
         {
           int i;
-          
 
-          for ( i = 0; /* No condition here */; i++ )
+	  validator = LAST_VALIDATE;
+          for ( i = 0; i < LAST_VALIDATE; i++ )
           {
-            if ( validator_symbols[i] == NULL )
-            {
-              fprintf( stderr, "*** Unknown validator name: %s\n", optarg );
-              print_usage();
-            }
-            
-            if ( strcmp( optarg, validator_symbols[i] ) == 0 )
+            if ( strcmp( optarg, validators[i].symbol ) == 0 )
             {
               validator = (ValidatorType)i;
               break;
             }
           }
+	  if ( validator == LAST_VALIDATE )
+	  {
+	    fprintf( stderr, "*** Unknown validator name: %s\n", optarg );
+	    print_usage( NULL );
+	  }
         }
         break;
 
@@ -654,12 +781,12 @@
         {
           fprintf( stderr, "*** Validation level is out of range: %d\n",
                            validation_level );
-          print_usage();
+          print_usage( NULL );
         }
         break;
 
       default:
-        print_usage();
+        print_usage( NULL );
         break;
       }
     }
@@ -670,12 +797,12 @@
     if ( argc == 0 )
     {
       fprintf(stderr, "*** Font file is not specified.\n");
-      print_usage();
+      print_usage( NULL );
     }
     else if ( argc > 1 )
     {
       fprintf(stderr, "*** Too many font files.\n");
-      print_usage();
+      print_usage( NULL );
     }
 
     fontfile = argv[0];
@@ -701,49 +828,31 @@
      * Run a validator
      */
     {
-      FT_Library  library;
       FT_Face     face;
-      FT_Error    error;
-
+      FT_Error    status;
 
       status = 0;
 
-      error = FT_Init_FreeType( &library );
-      if ( error )
-        panic ( error, "Could not initialize FreeType library" );
+      if ( !validators[validator].is_implemented( library ) )
+	panic( FT_Err_Unimplemented_Feature, 
+	       validators[validator].unimplemented_message );
+
 
       /* TODO: Multiple faces in a font file? */
       error = FT_New_Face( library, fontfile, 0, &face );
       if ( error )
         panic( error, "Could not open face." );
       
-      switch ( validator )
-      {
-      case OT_VALIDATE:
-        if ( dump_table_list == 0 )
-          status = run_ot_validator( face, tables, validation_level );
-        else
-          status = list_ot_tables  ( face );
-        break;
-      case GX_VALIDATE:
-        if ( dump_table_list == 0 )
-          status = run_gx_validator( face, tables, validation_level );
-        else
-          status = list_gx_tables( face );
-        break;
-      case CKERN_VALIDATE:
-        if ( dump_table_list == 0 )
-          status = run_ckern_validator( face, tables, validation_level );
-        else
-          status = list_ckern_tables( face );
-        break;
-      }
+      if ( dump_table_list )
+	validators[validator].list_tables( face );
+      else
+	status = validators[validator].run( face, tables, validation_level );
       
       FT_Done_Face( face );
       FT_Done_FreeType( library );
-    }
 
-    return status;
+      return (int)status;
+    }
   }
 
 
